(* seplog (c) AIST 2005-2015. R. Affeldt, N. Marti, et al. GNU GPLv3. *)
From mathcomp Require Import ssreflect ssrfun ssrbool eqtype seq.

(**********)

(* comparaison *)

Inductive cmp : Type :=
| LT
| EQ
| GT.

Module Compare.

Section Compare.

Variable A: Type.
Variable lt: A -> A -> Prop.

Variable compare: A -> A -> cmp.

CoInductive reflect_cmp (x y: A): cmp -> Set :=
| Reflect_LT: lt x y -> ~(x = y) -> ~(lt y x) -> reflect_cmp x y LT
| Reflect_EQ: ~(lt x y) -> x = y -> ~(lt y x) -> reflect_cmp x y EQ
| Reflect_GT: ~(lt x y) -> ~(x = y) -> lt y x -> reflect_cmp x y GT.

End Compare.

(*Check reflect_cmp.*)

Definition axiom T (lt: T -> T -> Prop) (e: T -> T -> cmp) := forall x y, reflect_cmp T lt x y (e x y).

Structure mixin_of T := Mixin {lt: T -> T -> Prop; op : T -> T -> cmp; _ : axiom T lt op}.
Notation class_of := mixin_of (only parsing).

Section ClassDef.

Structure type := Pack {sort; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.
Variables (T : Type) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.

Definition pack c := @Pack T c T.
Definition clone := fun c & cT -> T & phant_id (pack c) cT => pack c.

End ClassDef.

Module Exports.
Coercion sort : type >-> Sortclass.
Notation cmpType := type.
Notation CmpMixin := Mixin.
Notation CmpType T m := (@pack T m).
Notation "[ 'eqMixin' 'of' T ]" := (class _ : mixin_of T)
  (at level 0, format "[ 'eqMixin' 'of' T ]") : form_scope.
Notation "[ 'eqType' 'of' T 'for' C ]" := (@clone T C _ idfun id)
  (at level 0, format "[ 'eqType' 'of' T 'for' C ]") : form_scope.
Notation "[ 'eqType' 'of' T ]" := (@clone T _ _ id id)
  (at level 0, format "[ 'eqType' 'of' T ]") : form_scope.
End Exports.

End Compare.
Export Compare.Exports.

(*Print Compare.mixin_of.
Check Compare.op.
Check Equality.op.*)

Definition cmp_op T := Compare.op _ (Compare.class T).
Implicit Arguments cmp_op [T].
(*Check cmp_op.*)

Definition cmp_lt T := Compare.lt _ (Compare.class T).
(*Check cmp_lt.*)

Lemma cmpE : forall T x, @cmp_op T x = @Compare.op _ (Compare.class T) x.
done.
Qed.

Lemma cmpP : forall T, Compare.axiom _ (@cmp_lt T) (@cmp_op T).
case => s [] //=.
Qed.

Implicit Arguments cmpP [T x y].

Delimit Scope cmp_scope with CMP.
Open Scope cmp_scope.

Definition eq_cmp c1 c2 := match c1, c2 with | LT, LT => true | EQ, EQ => true | GT, GT => true | _, _ => false end.

Lemma eq_cmpP : Equality.axiom eq_cmp.
Proof.
case; case => //=; 
  match goal with
    | |- reflect _ true => apply ReflectT =>//=
    | |- reflect _ false => apply ReflectF =>//=
  end.
Qed.

Canonical Structure cmp_eqMixin := EqMixin eq_cmpP.
Canonical Structure cmp_eqType := Eval hnf in EqType _ cmp_eqMixin.


Definition cmp_cmp (c1 c2: cmp):=
  match c1, c2 with
    | LT, LT => EQ
    | LT, _ => LT
    | EQ, LT => GT
    | EQ, EQ => EQ
    | EQ, GT => LT
    | GT, GT => EQ
    | GT, _ => GT
  end.

Inductive lt_cmp: cmp -> cmp -> Prop :=
| lt_LT_EQ: lt_cmp LT EQ
| lt_LT_GT: lt_cmp LT GT
| lt_EQ_GT: lt_cmp EQ GT.

Lemma cmp_cmpP: Compare.axiom _ lt_cmp cmp_cmp.
case; case => //=;
match goal with
  | |- Compare.reflect_cmp _ _ _ _ EQ =>
    apply Compare.Reflect_EQ => //=; [ move => H; inversion H | move => H; inversion H ]
  | |- Compare.reflect_cmp _ _ _ _ LT =>
    apply Compare.Reflect_LT => //=; [ constructor | move => H; inversion H ]
  | |- Compare.reflect_cmp _ _ _ _ GT =>
    apply Compare.Reflect_GT => //=; [ move => H; inversion H | constructor ]
  | |- _ => idtac
end.
Qed.

Canonical Structure cmp_cmpMixin := CmpMixin _ _ _ cmp_cmpP.
Canonical Structure cmp_cmpType := Eval hnf in CmpType _ cmp_cmpMixin.

Notation "x < y" := (cmp_op x y == LT)
  (at level 70, no associativity) : bool_scope.

Notation "x <= y" := ((cmp_op x y == LT) || (cmp_op x y == EQ))
  (at level 70, no associativity) : bool_scope.

Notation "x > y" := (cmp_op x y == GT)
  (at level 70, no associativity) : bool_scope.

Notation "x >= y" := ((cmp_op x y == GT) || (cmp_op x y == EQ))
  (at level 70, no associativity) : bool_scope.

Notation "x === y" := (cmp_op x y == EQ)
  (at level 70, no associativity) : bool_scope.

(**********)

From mathcomp Require Import ssrnat.

Fixpoint cmp_nat (x y: nat):=
  match x, y with
    | O, O => EQ
    | S _, O => GT
    | O, S _ => LT
    | S x, S y => cmp_nat x y
  end.

Inductive lt_nat: nat -> nat -> Prop :=
| lt_O_S: forall x, lt_nat O (S x)
| lt_S_S: forall x y, lt_nat x y -> lt_nat (S x) (S y).

Lemma cmp_natP: Compare.axiom _ lt_nat cmp_nat.
red; induction x => //=.
- case.
  + constructor; [move => H; inversion H | done | move => H; inversion H].
  + move => n; constructor; [constructor | done | move => H; inversion H].
- case => //=.
  + constructor; [move => H; inversion H | done | constructor].
  + move => n.
    move: {IHx}(IHx n) => H; inversion H.
    * constructor => //=.
      - constructor => //=.
      - move/eqP=> //= H4; simpl in H4.
        apply: H2.
        apply/eqP.
        by rewrite -eqSS.
      - by move => H4; inversion H4; subst; clear H4.
    * constructor => //=.
      - by move => H4; inversion H4; subst; clear H4.
      - by apply/eqP; rewrite eqSS; apply/eqP.
      - by move => H4; inversion H4; subst; clear H4.
    * constructor => //=.
      - by move => H4; inversion H4; subst; clear H4.
      - move/eqP=> //= H4; simpl in H4.
        apply: H2.
        apply/eqP.
        by rewrite -eqSS.
      - constructor => //=.
Qed.

Canonical Structure nat_cmpMixin := CmpMixin _ _ _ cmp_natP.
Canonical Structure nat_cmpType := Eval hnf in CmpType _ nat_cmpMixin.
